## $Id: ARFA.pm,v 1.4 2006/05/20 21:36:54 ceratites Exp $
#
# BioPerl module for Bio::Tools::ARFA
#
# Automated Release Factor Annotation (ARFA)
# Copyright 2005-2006 Bekaert M <mbekaert@gmail.com>
#
# ARFA.pm is Open Source software. It's free for you to download and
# use as you wish. ARFA.pm's license is an Creative Commons
# Attribution-NonCommercial-ShareAlike 2.5 License. Read it if you aren't
# sure what you can or can't do.
# see: http://creativecommons.org/licenses/by-nc-sa/2.5/

# POD documentation - main docs before the code

=head1 NAME

Bio::Tools::ARFA - Automated Release Factor Annotation (ARFA)

=head1 SYNOPSIS

Take a sequence object from, say, an inputstream, and find bacterial
release factor prfA, prfB or prfH. HMM profile is used in order to
identify location, frame and orientation of such gene.

Creating the ARFA object, eg:

  my $inputstream = Bio::SeqIO->new(-file => 'seqfile', -format => 'fasta');
  my $seqobj = $inputstream->next_seq();
  my $ARFA_obj = Bio::Tools::ARFA->new($seqobj);

obtain an array holding the start point, the stop point, the DNA
sequence and amino-acid sequence, eg:

  my $array_ref = $ARFA_obj->{'_result'} if ($ARFA_obj->find());

display result in genbank format, eg:

  $ARFA_obj->show('genbank');

=head1 DESCRIPTION

Bio::Tools::ARFA is a featherweight object for identifying release
factor. Sequences should not be longer than 20kb. Returned array
include location, sequence and statistic for the putative release
factor. Initiation codon is not predicted precisely.

See Synopsis above for object creation code.

=head1 DRIVER SCRIPT

 #!/usr/bin/perl -w
 use strict;

 use Bio::Seq;
 use Bio::Tools::ARFA;

 my $inseq = Bio::SeqIO->new('-file' => "< $yourfile", -format => 'Fasta');
 while (my $seq = $inseq->next_seq) {
   my $RFF_obj = Bio::Tools::ARFA->new($seq,11);
   if ($RFF_obj->find()) {
     $RFF_obj->show('genbank');
   } else {
     print "  no hit!\n";
   }
 }

=head1 REQUIREMENTS

To use this module you need Bioperl (L<http://www.bioperl.org/>) modules,
HMMER distribution (L<http://hmmer.wustl.edu/>) and FASTA distribution
(L<ftp://ftp.ebi.ac.uk/pub/software/unix/fasta/>) if using whole genome search.

=head1 FEEDBACK

User feedback is an integral part of the evolution of this
modules. Send your comments and suggestions preferably
to author.

=head1 AUTHOR

B<Michael Bekaert> (mbekaert@gmail.com)

Address:
     Biosciences Institute, Recoding
     University College Cork
     Cork
     Ireland

=head1 SEE ALSO

perl(1), hmmpfam(1), fasta34(1), bioperl web site

=head1 LICENCE

Copyright 2005-2006 - Michael Bekaert

ARFA.pm is Open Source software. It's free for you to download and
use as you wish. ARFA.pm's license is an Creative Commons
Attribution-NonCommercial-ShareAlike 2.5 License. Read it if you aren't
sure what you can or can't do.

See: L<http://creativecommons.org/licenses/by-nc-sa/2.5/>

=head1 APPENDIX

The rest of the documentation details each of the object methods.
Internal methods are usually preceded with a _

=cut

package Bio::Tools::ARFA;
use strict;
use vars qw(@ISA $VERSION);
use Bio::DB::GenBank;
use Bio::Tools::HMMER::Results;
use Bio::SeqIO;
use Bio::Seq;
use Bio::SearchIO;
use Bio::PrimarySeq;
use Bio::PrimarySeqI;
use Bio::Tools::CodonTable;

$VERSION = '1.4';
@ISA     = qw(Bio::Root::Root Bio::Root::IO);

# Default path
my $PATH_HMMPFAM = '/usr/local/bin/hmmpfam';
my $PATH_FASTA   = '/usr/local/bin/tfasta34';
my $PATH_REF     = './ARFA.fasta';
my $PATH_HMM     = './ARFA.hmm';
my $PATH_FS      = './ARFA.fs.hmm';
my $PATH_TMP     = '/tmp';

=head2 new

 Title   : new
 Usage   : my $nb = Bio::Tools::ARFA->new($seqobj,$table,$aug,$hmm,$fs,$pfam);
 Function: Initialize ARFA object.
 Returns : RFFinder object.
 Args    : $seqobj (mandatory) PrimarySeqI object (dna or rna),
           $table (optional) translation table number (default 1),
           $aug (optional) force AUG as start codon (default 0),
           $hmm (optional) path to hmm profiles by default ARFA
           look at ./ARFA.hmm,
           $fs (optional) path to frameshift hmm profiles by default
           ARFA look at ./ARFA.fs.hmm,
           $pfam (optional) path to hmmpfam from HMMER distribution
           by default ARFA look at /usr/local/bin/hmmpfam.

=cut

sub new {
    my ( $self, @args ) = @_;
    my ( $seqobj, $table, $aug, $hmm, $fs, $pfam ) = @args;
    $self = {};

    $self->{'_aug'} = ( ( defined $aug && int($aug) == 1 ) ? 1    : 0 );
    $self->{'_hmm'} = ( ( defined $hmm )                   ? $hmm : $PATH_HMM );
    $self->{'_fs'}  = ( ( defined $fs )                    ? $fs  : $PATH_FS );
    $self->{'_pfam'} = ( ( defined $pfam ) ? $pfam : $PATH_HMMPFAM );
    $seqobj->throw(
        "die in _initialize, ARFA works only on PrimarySeqI objects\n")
      unless ( $seqobj->isa("Bio::PrimarySeqI") );
    $seqobj->throw("die in _initialize, ARFA works only on DNA sequences\n")
      if ( $seqobj->alphabet eq 'protein' );
    $seqobj->throw(
        "die in _initialize, ARFA works only on DNA sequences < 20kb\n")
      if ( length( $seqobj->seq() ) > 20000 );
    $seqobj->throw("die in _initialize, hmmpfam not found\n")
      unless ( -f $self->{'_pfam'} );
    $seqobj->throw("die in _initialize, hmm profile not found\n")
      unless ( -f $self->{'_hmm'} );
    $seqobj->throw("die in _initialize, hmm profile not found\n")
      unless ( -f $self->{'_fs'} );
    my $chrom =
      ( ( defined $seqobj->display_id ) ? $seqobj->display_id : 'noname' );
    $seqobj = uc $seqobj->seq();
    $seqobj =~ tr/U/T/;
    $self->{'_seqref'} =
      Bio::Seq->new( -seq => $seqobj, -alphabet => 'dna', -id => $chrom );
    $self->{'_table'} = ( ( defined $table ) ? $table : 1 );
    bless($self);
    return $self;
}

=head2 find

 Title   : find
 Usage   : my $bool = $ARFA_obj->find($strand,$start,$end);
 Function: Identify release factor.
 Returns : boolean if release factor found.
 Args    : $evalue (optional) Set the E-value threshold,
           $strand (optional) Strand where search should be done
           (1 direct, -1 reverse or 0 both). Default is 0,
           $start (optional) coordinate of the first nucleotide. Useful
           for coordinate calculation if first is not 1. Default is 1,
           $end (optional) coordinate of the first nucleotide. Useful
           for coordinate calculation if last is not 1. Default is length.

=cut

sub find {
    my ( $self, @args ) = @_;
    my ( $evalue, $strand, $start, $end ) = @args;
    $self->{'_evalue'} =
      ( ( defined $evalue ) && ( $evalue > 0 ) ) ? $evalue : 1e-40;
    $strand = 0
      unless ( ( defined $strand ) && ( $strand == 1 || $strand == -1 ) );
    $start = 1 unless ( ( defined $start ) && $start > 1 );
    $end = length( $self->{'_seqref'}->seq() )
      unless ( ( defined $end ) && $end > 1 && $end > $start );
    my $status = $self->_what_rf($strand);
    if ( ($status) && ( $self->{'_hmmrf'}[0] < $self->{'_evalue'} ) ) {
        print '> ', $self->{'_hmmrf'}[1], ' found (', $self->{'_hmmrf'}[0],
          ') in frame ', $self->{'_hmmrf'}[2],
          (
            (
                     ( defined $self->{'_hmmfs'} )
                  && ( $self->{'_hmmrf'}[1] eq 'RF2' )
                  && ( $self->{'_hmmrf'}[2] != $self->{'_hmmfs'}[2] )
            ) ? ( ' and first ORF in frame ' . $self->{'_hmmfs'}[2] ) : ''
          ),
          "\n"
          if ( $self->{'_verbose'} );
        return $self->_find_orf( $self->{'_hmmrf'}[6], $start, $end );
    }
    return 0;
}

=head2 _what_rf

 Title   : _what_rf
 Usage   : my $bool = $ARFA_obj->_what_rf($strand);
 Function: Use HMM profile to identify release factor.
 Returns : boolean.
 Args    : $strand (mandatory) Strand where search should be done
           (1 direct, -1 reverse or 0 both).

=cut

sub _what_rf {
    my ( $self, @args ) = @_;
    my $strand = shift(@args);

    my $seq = $self->{'_seqref'};
    $seq = $seq->revcom if ( $strand < 0 );
    $self->{'_file'} = 'rf_bot.seq' if ( !defined $self->{'_file'} );
    open SEQ, '>' . $PATH_TMP . '/' . $self->{'_file'};
    for ( my $i = 0 ; $i < 3 ; $i++ ) {
        print SEQ ">$i\n",
          $seq->translate( undef, undef, $i, $self->{'_table'} )->seq(), "\n";
    }
    if ( $strand == 0 ) {
        $seq = $seq->revcom;
        for ( my $i = 0 ; $i < 3 ; $i++ ) {
            print SEQ ">$i-\n",
              $seq->translate( undef, undef, $i, $self->{'_table'} )->seq(),
              "\n";
        }
    }
    close SEQ;
    system( $self->{'_pfam'} . ' '
          . $self->{'_hmm'} . ' '
          . $PATH_TMP . '/'
          . $self->{'_file'} . ' > '
          . $PATH_TMP . '/'
          . $self->{'_file'}
          . '.report' );
    my $hmmer = new Bio::Tools::HMMER::Results(
        -file => $PATH_TMP . '/' . $self->{'_file'} . '.report',
        -type => 'hmmpfam'
    );
    foreach my $domain ( $hmmer->each_Domain ) {
        $self->{'_hmmrf'} = (
            [
                $domain->evalue,
                $domain->hmmname,
                substr( $domain->seq_id, 0, 1 ),
                $domain->bits,
                $domain->start,
                $domain->end,
                (
                    ( $strand != 0 )
                    ? $strand
                    : ( ( length( $domain->seq_id ) > 1 ) ? -1 : 1 )
                )
            ]
          )
          if (
            (
                ( !defined $self->{'_hmmrf'}[0] )
                || $domain->evalue < $self->{'_hmmrf'}[0]
            )
            && ( $domain->hmmname ne 'RF2_ORF0' )
          );
        $self->{'_hmmfs'} = (
            [
                $domain->evalue,
                $domain->hmmname,
                substr( $domain->seq_id, 0, 1 ),
                $domain->bits,
                $domain->start,
                $domain->end,
                (
                    ( $strand != 0 )
                    ? $strand
                    : ( ( length( $domain->seq_id ) > 1 ) ? -1 : 1 )
                )
            ]
          )
          if (
            (
                ( !defined $self->{'_hmmfs'}[0] )
                || $domain->evalue < $self->{'_hmmfs'}[0]
            )
            && ( $domain->hmmname eq 'RF2_ORF0' )
          );
    }
    unlink( $PATH_TMP . '/' . $self->file );
    unlink( $PATH_TMP . '/' . $self->file . '.report' );
    return ( defined $self->{'_hmmrf'}[0] ) ? 1 : 0;
}

=head2 _what_fs

 Title   : _what_fs
 Usage   : my $evalue = $ARFA_obj->_what_rs($site);
 Function: Use HMM profile to qualify -1 frameshift site.
 Returns : double (e-value).
 Args    : $site (mandatory) Frameshift site.

=cut

sub _what_fs {
    my ( $self, @args ) = @_;
    my $site = shift(@args);
    my $evalue;

    $self->{'_file'} = 'rf_bot.seq' if ( !defined $self->{'_file'} );
    open SEQ, '>' . $PATH_TMP . '/' . $self->{'_file'};
    print SEQ ">fs\n", $site, "\n";
    close SEQ;
    system( $self->{'_pfam'} . ' -n '
          . $self->{'_fs'} . ' '
          . $PATH_TMP . '/'
          . $self->{'_file'} . ' > '
          . $PATH_TMP . '/'
          . $self->{'_file'}
          . '.report' );
    my $hmmer = new Bio::Tools::HMMER::Results(
        -file => $PATH_TMP . '/' . $self->{'_file'} . '.report',
        -type => 'hmmpfam'
    );

    foreach my $domain ( $hmmer->each_Domain ) {
        $evalue = $domain->evalue if ( $domain->hmmname eq 'fs' );
    }
    unlink( $PATH_TMP . '/' . $self->file );
    unlink( $PATH_TMP . '/' . $self->file . '.report' );
    return $evalue;
}

=head2 _find_orf

 Title   : _find_orf
 Usage   : my $bool = $ARFA_obj->_find_orf($strand,$start,$end);
 Function: Retreave ORF release factor.
 Returns : boolean.
 Args    : $strand (mandatory) Strand where RF have been found
           (1 direct or -1 reverse),
           $start (mandatory) coordinate of the first nucleotide,
           $end (mandatory) coordinate of the last nucleotide.

=cut

sub _find_orf {
    my ( $self, @args ) = @_;
    my ( $strand, $start, $end ) = @args;
    my ( $position1, $position2 );
    my ( $begin, $stop ) = $self->_translation();
    my $seq = $self->{'_seqref'};
    $seq = $seq->revcom if ( $strand < 0 );
    $seq = $seq->seq();
    if (   ( defined $self->{'_hmmrf'} )
        && ( defined $self->{'_hmmfs'} )
        && ( $self->{'_hmmrf'}[1] eq 'RF2' )
        && ( $self->{'_hmmrf'}[2] == ( $self->{'_hmmfs'}[2] + 1 ) % 3 ) )
    {

        #RF2 with +1FS
        my $mydna =
          substr( $seq, 0,
            ( $self->{'_hmmfs'}[4] * 3 + abs( $self->{'_hmmfs'}[2] ) ) );
        if ( $mydna =~ m/(($begin)((?!($stop|$begin))(.{3}))*?)$/o ) {
            $position1 = length($`);
        }
        else {
            if ( $mydna =~ m/(($stop)((?!($stop))(.{3}))*)$/o ) {
                $position1 = length($`) + 3;
                $self->{'_noatg'} = 1;
            }
        }
        if ( defined $position1 ) {
            my $mydna = substr( $seq, $position1 );
            if ( $mydna =~
m/^(.{3})(((?!($stop))(.{3}))*)($stop)(.((?!($stop))(.{3})){75,})($stop)/o
              )
            {
                my ( $position3, $coord );
                if ( $strand > 0 ) {
                    $position2 =
                      $start + $position1 + length( $1 . $2 . $6 . $7 . $11 ) -
                      1;
                    $position3 = $start + $position1 + length( $1 . $2 );
                    $position1 = $start + $position1;
                    $coord     = 'join('
                      . $position1 . '..'
                      . ( $position3 - 1 ) . ','
                      . ( $position3 + 1 ) . '..'
                      . $position2 . ')';
                }
                else {
                    $position2 = $end - $position1;
                    $position3 = $end - $position1 - length( $1 . $2 );
                    $position1 =
                      $end - $position1 - length( $1 . $2 . $6 . $7 . $11 ) + 1;
                    $coord =
                        'complement(join('
                      . $position1 . '..'
                      . ( $position3 - 1 ) . ','
                      . ( $position3 + 1 ) . '..'
                      . $position2 . '))';
                }
                $self->{'_result'} = (
                    [
                        $coord,
                        $position1,
                        $position2,
                        $strand,
                        'Bacterial release factor 2',
                        'prfB',
                        (
                            (
                                ( defined $self->{'_noatg'} )
                                ? 'The position of the initiation codon is not identified; '
                                : ''
                            )
                            . 'RF2 ORF0: '
                              . $self->{'_hmmfs'}[0]
                              . ' ORF1: '
                              . $self->{'_hmmrf'}[0]
                        ),
                        $1 . $2 . $6 . $7 . $11,
                        Bio::Seq->new(
                            -seq     => $1 . $2 . substr( $6, 1 ) . $7,
                            alphabet => 'dna'
                          )->translate( undef, undef, undef, $self->{'_table'} )
                          ->seq()
                    ]
                );
            }
        }
    }
    elsif ( defined $self->{'_hmmrf'} ) {

        #RF2 without +1FS / RF1 / RF0
        my $mydna =
          substr( $seq, 0,
            ( $self->{'_hmmrf'}[4] * 3 + abs( $self->{'_hmmrf'}[2] ) + 9 ) );
        if ( $mydna =~ m/(($begin)((?!($stop|$begin))(.{3}))*?)$/o ) {
            $position1 = length($`);
        }
        else {
            if ( $mydna =~ m/(($stop)((?!($stop))(.{3}))*)$/o ) {
                $position1 = length($`) + 3;
                $self->{'_noatg'} = 1;
            }
        }
        if ( defined $position1 ) {
            my $mydna = substr( $seq, $position1 );
            my ( $coord, $dna, $stopcodon );
            if ( $mydna =~ m/^(.{3})(((?!($stop))(.{3})){75,})($stop)/o ) {
                $dna       = $1 . $2;
                $stopcodon = $6;
            }
            elsif ( $seq =~ m/($stop)(((?!($stop))(.{3})){250,})($stop)/o ) {
                $self->{'_noatg'} = 1;
                $position1        = length($`) + 3;
                $dna              = $2;
                $stopcodon        = $6;
            }
            if ( defined $dna ) {
                if ( $strand > 0 ) {
                    $position2 =
                      $start + $position1 + length( $dna . $stopcodon ) - 1;
                    $position1 = $start + $position1;
                    $coord     = $position1 . '..' . $position2;
                }
                else {
                    $position2 = $end - $position1;
                    $position1 =
                      $end - $position1 - length( $dna . $stopcodon ) + 1;
                    $coord =
                      'complement(' . $position1 . '..' . $position2 . ')';
                }
                $self->{'_result'} = (
                    [
                        $coord,
                        $position1,
                        $position2,
                        $strand,
                        'Bacterial release factor '
                          . (
                            ( $self->{'_hmmrf'}[1] eq 'RF1' ) ? '1'
                            : (
                                ( $self->{'_hmmrf'}[1] eq 'RF3' ) ? '3'
                                : ( ( $self->{'_hmmrf'}[1] eq 'RFH' ) ? 'H'
                                    : '2' )
                            )
                          ),
                        (
                            ( $self->{'_hmmrf'}[1] eq 'RF1' ) ? 'prfA'
                            : (
                                ( $self->{'_hmmrf'}[1] eq 'RF3' ) ? 'prfC'
                                : ( ( $self->{'_hmmrf'}[1] eq 'RFH' ) ? 'prfH'
                                    : 'prfB' )
                            )
                        ),
                        (
                            (
                                ( defined $self->{'_noatg'} )
                                ? 'The position of the initiation codon is not identified; '
                                : ''
                            )
                            . $self->{'_hmmrf'}[1]
                              . ' ORF: '
                              . $self->{'_hmmrf'}[0]
                        ),
                        $dna . $stopcodon,
                        Bio::Seq->new( -seq => $dna, alphabet => 'dna' )
                          ->translate( undef, undef, undef, $self->{'_table'} )
                          ->seq()
                    ]
                  )
                  if ( defined $coord );
            }
        }
    }
    return ( defined $position2 ) ? 1 : 0;
}

=head2 show

 Title   : show
 Usage   : $ARFA_obj->show($outstyle);
 Function: Print result in various style.
 Returns : none.
 Args    : $outstyle (mandatory) Genbank or XML style.

=cut

sub show {
    my ( $self, @args ) = @_;
    my $out = shift(@args);
    if ( $out eq 'genbank' ) {
        return undef unless ( defined $self->{'_result'} );
        print '     gene            ',
          (
            ( $self->{'_result'}[3] < 0 )
            ? ( 'complement('
                  . $self->{'_result'}[1] . '..'
                  . $self->{'_result'}[2]
                  . ')' )
            : ( $self->{'_result'}[1] . '..' . $self->{'_result'}[2] )
          ),
          "\n";
        print '                     /locus_tag="',
          $self->{'_seqref'}->display_id, "\"\n";
        print '                     /gene="', $self->{'_result'}[5], "\"\n";

        #      my $dna = $self->{'_result'}[7].'"';
        #      print '                     /dna="',substr($dna,0,52),"\n";
        my $i = 0;

    #      while (length($dna)>(($i*58)+52)) {
    #         print '                     ',substr($dna,(($i++*58)+52),58),"\n";
    #      }
        print '     CDS             ',        $self->{'_result'}[0], "\n";
        print '                     /gene="', $self->{'_result'}[5], "\"\n";
        print '                     /locus_tag="',
          $self->{'_seqref'}->display_id, "\"\n";
        print '                     /note="', $self->{'_result'}[6], "\"\n";
        if ( $self->{'_result'}[6] =~ /ORF1/ ) {
            my $position3 = ( $1 + 1 )
              if ( $self->{'_result'}[0] =~ /^\D*\d+\.\.(\d+)/ );
            my $fs = substr(
                $self->{'_result'}[7],
                (
                      ( $self->{'_result'}[3] > 0 )
                    ? ( $position3 - $self->{'_result'}[1] - 15 )
                    : ( $self->{'_result'}[2] - $position3 - 15 )
                ),
                19
            );
            print '                     /inference="FS site: ',
              $self->_what_fs($fs), "\"\n";
        }
        print "                     /codon_start=1\n"
          unless ( defined $self->{'_noatg'} );
        print '                     /transl_table=', $self->{'_table'}, "\n";
        print '                     /product="', $self->{'_result'}[4], "\"\n";
        my $translation_issue = $self->{'_result'}[8] . '"';
        print '                     /translation="',
          substr( $translation_issue, 0, 44 ), "\n";

        #      $i=0;
        while ( length($translation_issue) > ( ( $i * 58 ) + 44 ) ) {
            print '                     ',
              substr( $translation_issue, ( ( $i++ * 58 ) + 44 ), 58 ), "\n";
        }
    }
    elsif ( $out eq 'xml-begin' ) {
        print "<arfaxml version=\"1.1\">\n";
        print " <analysis>\n";
        print "  <program>\n";
        print "   <prog-name>ARFA.pm</prog-name>\n";
        print "   <prog-version>$VERSION</prog-version>\n";
        print "  </program>\n";
        print "  <date>\n";
        print '   <day>', (gmtime)[3], "</day>\n";
        print '   <month>', (gmtime)[4] + 1,    "</month>\n";
        print '   <year>',  (gmtime)[5] + 1900, "</year>\n";
        print "  </date>\n";
        print "  <parameter>\n";
        print '   <evalue>', shift(@args), "</evalue>\n";
        print '   <table>',  shift(@args), "</table>\n";
        print "  </parameter>\n";
        print " </analysis>\n";
    }
    elsif ( $out eq 'xml-end' ) {
        print "</arfaxml>\n";
    }
    elsif ( $out eq 'xml' ) {
        return undef unless ( defined $self->{'_result'} );
        my $organism = shift(@args);
        print ' <sequence id="', $self->{'_seqref'}->display_id, ".seq\">\n";
        print "  <input>\n";
        print '   <organism genbank="'
          . $self->{'_seqref'}->display_id
          . "\">$organism</organism>\n"
          if ( defined $organism );
        print '   <seq type="dna" length="', length( $self->{'_seqref'}->seq ),
          '">', $self->{'_seqref'}->seq, "</seq>\n";
        print "  </input>\n";
        print '  <output id="', $self->{'_seqref'}->display_id, "\">\n";
        print '   <gene id="',  $self->{'_seqref'}->display_id, ".1\">\n";
        print '    <coord>',
          (
            ( $self->{'_result'}[3] < 0 )
            ? ( 'complement('
                  . $self->{'_result'}[1] . '..'
                  . $self->{'_result'}[2]
                  . ')' )
            : ( $self->{'_result'}[1] . '..' . $self->{'_result'}[2] )
          ),
          "</coord>\n";
        print '    <name>', $self->{'_result'}[5], "</name>\n";
        print '    <seq type="dna" length="', length( $self->{'_result'}[7] ),
          '">', $self->{'_result'}[7], "</seq>\n";
        print "   </gene>\n";
        print '   <cds id="', $self->{'_seqref'}->display_id, ".2\">\n";
        print '    <coord',
          ( defined $self->{'_noatg'} ) ? ' start="unknown"' : '', '>',
          $self->{'_result'}[0], "</coord>\n";
        print '    <name>',    $self->{'_result'}[5], "</name>\n";
        print '    <note>',    $self->{'_result'}[6], "</note>\n";
        print '    <product>', $self->{'_result'}[4], "</product>\n";
        print '    <seq type="prt" length="', length( $self->{'_result'}[8] ),
          '">', $self->{'_result'}[8], "</seq>\n";
        print '    <model hmm="', $self->{'_hmmrf'}[1], '">',
          $self->{'_hmmrf'}[0], "</model>\n";
        print '    <model hmm="', $self->{'_hmmfs'}[1], '">',
          $self->{'_hmmfs'}[0], "</model>\n"
          if ( $self->{'_result'}[6] =~ /ORF1/ );
        print "   </cds>\n";

        if ( $self->{'_result'}[6] =~ /ORF1/ ) {
            my $position3 = ( $1 + 1 )
              if ( $self->{'_result'}[0] =~ /^\D*\d+\.\.(\d+)/ );
            my $fs = substr(
                $self->{'_result'}[7],
                (
                      ( $self->{'_result'}[3] > 0 )
                    ? ( $position3 - $self->{'_result'}[1] - 15 )
                    : ( $self->{'_result'}[2] - $position3 - 15 )
                ),
                19
            );
            print '   <frameshift id="', $self->{'_seqref'}->display_id,
              '.3" evalue="', $self->_what_fs($fs), "\">\n";
            my $sd = $self->_sd_align( 'GTGGAGGAA', substr( $fs, 0, 12 ), 1 );
            print '    <sd spacer="', length($2), '">', $sd, "</sd>\n"
              if ( $fs =~ /($sd)(.*?).{7}$/ );
            print '    <seq type="rna" length="', 7 + length( $2 . $1 ), '">',
              substr( $fs, 12 - length( $2 . $1 ) ), "</seq>\n";
            print '    <psite>',      substr( $fs, 12, 3 ), "</psite>\n";
            print '    <asite>',      substr( $fs, 15, 3 ), "</asite>\n";
            print '    <downstream>', substr( $fs, 18, 1 ), "</downstream>\n";
            print "   </frameshift>\n";
        }
        print "  </output>\n";
        print " </sequence>\n";
    }
}

=head2 _sd_align

 Title   : _sd_align
 Usage   : my $sd = $ARFA_obj->_sd_align($a,$b,$mis);
 Function: Find SD sequence from mRNA sequence (this dynamic algorithm
           allows mismathing but not gap between SD and anti-SD
           sequence).
 Returns : $sd Shine-Dalgarno sequence.
 Args    : $a (mandatory) sequence of 16S rRNA,
           $b (mandatory) sequence of the mRNA,
           $mis (optional) allows mismatches if set.

=cut

sub _sd_align {
    my ( $self, @args ) = @_;
    my ( $a, $b, $mis ) = @args;
    my $S = [];
    my $n = length($a);
    my $m = length($b);
    my @a = split // => uc $a;
    my @b = split // => uc $b;
    my ( $zi, $zj, $zz ) = ( 1, 1, 0 );
    for ( my $ii = 0 ; $ii <= $n ; ++$ii ) { $S->[$ii][0] = 0; }
    for ( my $jj = 0 ; $jj <= $m ; ++$jj ) { $S->[0][$jj] = 0; }

    for ( my $ii = 1 ; $ii <= $n ; ++$ii ) {
        for ( my $jj = 1 ; $jj <= $m ; ++$jj ) {
            $S->[$ii][$jj] =
              (      ( $a[ $ii - 1 ] eq $b[ $jj - 1 ] )
                  || ( $a[ $ii - 1 ] eq 'G' && $b[ $jj - 1 ] eq 'A' )
                  || ( $a[ $ii - 1 ] eq 'A' && $b[ $jj - 1 ] eq 'G' ) )
              ? $S->[ $ii - 1 ][ $jj - 1 ] + 1
              : ( ( $S->[ $ii - 1 ][ $jj - 1 ] > 1 && defined $mis )
                ? $S->[ $ii - 1 ][ $jj - 1 ] - 1
                : 0 );
            if ( $zz < $S->[$ii][$jj] ) {
                $zi = $ii;
                $zj = $jj;
                $zz = $S->[$ii][$jj];
            }
        }
    }
    my $ii  = $zi;
    my $jj  = $zj;
    my $zz2 = '';
    my $lcs = '';
    while ( $ii > 0 && $jj > 0 && $S->[ $ii-- ][$jj] > 0 ) {
        $lcs = $b[ --$jj ] . $lcs;
    }
    return $lcs;
}

=head2 _translation

 Title   : _translation
 Usage   : my ($start,$end) = $ARFA_obj->_translation();
 Function: format initiation and stop codons for regex.
 Returns : array with initiation and stop codons.
 Args    : none.

=cut

sub _translation {
    my $self = shift;

    my @table        = ( 'A', 'T', 'C', 'G' );
    my @var          = ();
    my @var2         = ();
    my $var_i        = 0;
    my $var2_i       = 0;
    my $myCodonTable = Bio::Tools::CodonTable->new( -id => $self->{'_table'} );
    for ( my $i = 0 ; $i < 4 ; $i++ ) {
        for ( my $j = 0 ; $j < 4 ; $j++ ) {
            for ( my $k = 0 ; $k < 4 ; $k++ ) {
                $var[ $var_i++ ] = $table[$i] . $table[$j] . $table[$k]
                  if $myCodonTable->is_start_codon(
                    $table[$i] . $table[$j] . $table[$k] );
                $var2[ $var2_i++ ] = $table[$i] . $table[$j] . $table[$k]
                  if $myCodonTable->is_ter_codon(
                    $table[$i] . $table[$j] . $table[$k] );
            }
        }
    }
    @var = ('ATG') if ( $self->{'_aug'} );
    return ( join( '|', @var ), join( '|', @var2 ) );
}

=head2 getHits

 Title   : getHits
 Usage   : my @hits = Bio::Tools::ARFA->getHits($evalue,$seq,$ref,$fasta);
 Function: Quick localization of release factor (use FASTA).
 Returns : Array of hits
 Args    : $evalue (mandatory) Set the E-value threshold,
           $seq (mandatory) PrimarySeqI object (dna),
           $ref (optional) path to fasta reference file by default ARFA
           look at ./ARFA.fasta,
           $fasta (optional) path to tfasta34 from FASTA distribution
           by default ARFA look at /usr/local/bin/tfasta34.

=cut

sub getHits {
    my ( $self, @args ) = @_;
    my ( $seq, $evalue, $ref, $fasta ) = @args;
    $fasta = ( ( defined $fasta ) ? $fasta : $PATH_FASTA );
    $ref   = ( ( defined $ref )   ? $ref   : $PATH_REF );
    $evalue = ( ( defined $evalue ) && ( $evalue > 0 ) ) ? $evalue : 1;
    my @hits;
    if ( defined $seq ) {
        open( TMP, ">$PATH_TMP/genbank.seq" );
        print TMP ">query\n", $seq->seq(), "\n\n";
        close(TMP);
        system( $fasta
              . ' -Q -b 1 -d 1 -H '
              . $ref
              . " $PATH_TMP/genbank.seq> $PATH_TMP/genbank.report" );
        my $in = new Bio::SearchIO(
            -format => 'fasta',
            -file   => "$PATH_TMP/genbank.report"
        );
        eval {
            while ( my $result = $in->next_result )
            {
                while ( my $hit = $result->next_hit ) {
                    while ( my $hsp = $hit->next_hsp ) {
                        push(
                            @hits,
                            (
                                    $hsp->start('hit') . '|'
                                  . $hsp->end('hit') . '|'
                                  . $hit->strand('query')
                            )
                          )
                          if ( $hsp->evalue <= $evalue );
                    }
                }
            }
        };
        unlink("$PATH_TMP/genbank.report");
        unlink("$PATH_TMP/genbank.seq");
    }
    if ( $#hits >= 0 ) {
        @hits = sort @hits;
        my @hits2 = ();
        for ( my $i = 0 ; $i < $#hits ; $i++ ) {
            my ( $hitstart, $hitend, $hitstrand ) = split( /\|/, $hits[$i] );
            my $loop = 0;
            do {
                $loop = 0;
                my ( $hitstart_n, $hitend_n, $hitstrand_n ) =
                  split( /\|/, $hits[ $i + 1 ] )
                  if ( defined $hits[ $i + 1 ] );

                #frameshift or echo ?
                if (
                    ( defined $hits[ $i + 1 ] )
                    && (   ( abs( $hitstart_n - $hitstart ) < 1500 )
                        || ( abs( $hitend_n - $hitend ) < 1500 )
                        || abs( $hitstart_n - $hitend ) < 1500 )
                    && ( $hitstrand == $hitstrand_n )
                  )
                {
                    $i++;
                    $hitend   = $hitend_n   if ( $hitend < $hitend_n );
                    $hitstart = $hitstart_n if ( $hitstart > $hitstart_n );
                    $loop     = 1;
                }
            } until ( $loop == 0 );
            push( @hits2, ( $hitstart . '|' . $hitend . '|' . $hitstrand ) );
        }
        @hits = @hits2;
    }
    return @hits;
}

=head2 get_sequence

 Title   : get_sequence
 Usage   : my @array = Bio::Tools::ARFA->get_sequence($reference);
 Function: Download sequence from Genbank safely.
 Returns : $table translation table,
           $organism organism common name,
           $seqobj PrimarySeqI objects.
 Args    : $reference (mandatory) Genbank accession number.

=cut

sub get_sequence {
    my ( $self, @args ) = @_;
    my $reference = shift(@args);
    my ( $seq, $translate, $organism );
    my $loop = 1;
    until ( ( defined $seq ) || ( $loop++ > 3 ) ) {
        my $gb = new Bio::DB::GenBank(
            -retrievaltype => 'tempfile',
            -format        => 'genbank'
        );
        eval {
            my $seqio = $gb->get_Stream_by_id($reference);
            $seq = $seqio->next_seq;
        };
    }
    if ( defined $seq ) {
        foreach my $feat ( $seq->all_SeqFeatures() ) {
            $translate = 1;
            if ( $feat->primary_tag eq 'source' ) {
                $organism =
                  ( $feat->has_tag('organism')
                    ? ( $feat->each_tag_value('organism') )[0]
                    : '' );
            }
            elsif ( $feat->primary_tag eq 'CDS' ) {
                $translate = ( $feat->each_tag_value('transl_table') )[0]
                  if ( $feat->has_tag('transl_table') );
                last;
            }
        }
        return ( $translate, $organism, $seq );
    }
    return undef;
}

# and that's all the module
1;
