#!/usr/bin/perl -w
#
# Automated Release Factor Annotation (ARFA)
# Copyright 2005-2006 Bekaert M <m.bekaert@ucc.ie>
#
# rf_bot.pl is Open Source software. It's free for you to download and
# use as you wish. rf_bot.pl's license is an Creative Commons
# Attribution-NonCommercial-ShareAlike 2.5 License. Read it if you aren't
# sure what you can or can't do.
# see: http://creativecommons.org/licenses/by-nc-sa/2.5/

use strict;
use Getopt::Long;
use File::Basename;
use Cwd;
use Bio::Seq;

use Bio::Tools::ARFA;

#----------------------------------------
my $version = '1.4';
my $hmm     = getcwd() . '/rf_bot.hmm';
my $fs      = getcwd() . '/rf_bot.fs.hmm';
my $ref     = getcwd() . '/rf_bot.fasta';

#----------------------------------------------------------

sub hmm_disc {
    my (
        $seq, $name, $infile, $id, $translate, $evalue,
        $xml, $aug,  $hmm,    $fs, $pfam,      $verbose
      )
      = @_;
    my $chrom;
    if ( defined $name ) {
        $chrom = $name . '_' . $id;
    }
    elsif ( defined $seq->display_id ) {
        $chrom = $seq->display_id;
    }
    else {
        $chrom = 'noname_' . $id;
    }
    $seq->display_id($chrom);
    my $RFF_obj =
      Bio::Tools::ARFA->new( $seq, $translate, $aug, $hmm, $fs, $pfam );
    $RFF_obj->{'_verbose'} = 1 if ($verbose);
    $RFF_obj->{'_file'} = basename($infile);
    if ( $RFF_obj->find($evalue) ) {
        $RFF_obj->show( ($xml) ? 'xml' : 'genbank' );
    }
    else {
        print "  no hit!\n" unless ($xml);
    }
}

sub fasta_filter {
    my (
        $seq,       $ref,    $fasta, $organism, $id,
        $translate, $evalue, $xml,   $aug,      $hmm,
        $fs,        $pfam,   $verbose
      )
      = @_;
    my @hits = Bio::Tools::ARFA->getHits( $seq, 1, $ref, $fasta );
    if ( $#hits >= 0 ) {
        for ( my $i = 0 ; $i <= $#hits ; $i++ ) {
            my ( $hitstart, $hitend, $hitstrand ) = split( /\|/, $hits[$i] );
            my $seqstart = $hitstart - 350;
            $seqstart = 1 if ( $seqstart < 1 );
            my $seqend = $hitend + 350;
            $seqend = $seq->length() if ( $seqend > $seq->length() );
            my $seq_rf = Bio::Seq->new(
                -seq      => $seq->subseq( $seqstart, $seqend ),
                -alphabet => 'dna',
                -id       => $seq->display_id . '|' . $id
            );
            my $RFF_obj =
              Bio::Tools::ARFA->new( $seq_rf, $translate, $aug, $hmm, $fs,
                $pfam );
            $RFF_obj->{'_verbose'} = 1 if ($verbose);

            if ( $RFF_obj->find( $evalue, $hitstrand, $seqstart, $seqend ) ) {
                $RFF_obj->show( ($xml) ? 'xml' : 'genbank', $organism );
            }
            else {
                print "  no hit!\n" unless ($xml);
            }
        }
    }
}

#------------------------ Main ----------------------------
my ( $infile, $genbank, $translate, $name, $aug );
my $evalue  = 1e-40;
my $verbose = 0;
my $xml     = 0;
my $out     = 1;
GetOptions(
    'v!'        => \$verbose,
    'in:s'      => \$infile,
    'genbank:s' => \$genbank,
    'table:i'   => \$translate,
    'name:s'    => \$name,
    'e:f'       => \$evalue,
    'x!'        => \$xml,
    'f!'        => \$aug
);

if (
    (
           ( ( defined $genbank ) && !( defined $infile ) )
        || ( !( defined $genbank ) && ( defined $infile ) && ( -f $infile ) )
    )
    && ( !( defined $translate )
        || (   ( defined $translate )
            && ( $translate =~ /([1-9]|1[1-6]|2(1|2))/ ) ) )
  )
{
    $translate = 11
      unless ( defined $translate );    #We are probably looking for Bacteria...
    print "\n..:: Automated Release Factor Annotation (ARFA) ::..\n> Standalone program version $version <\n\n"
      if ($verbose);
    if ( defined $infile ) {
        if (
            (
                my $inseq =
                Bio::SeqIO->new( '-file' => "<$infile", '-format' => 'fasta' )
            )
          )
        {
            $out--;
            my $id = 0;
            Bio::Tools::ARFA->show( 'xml-begin', $evalue, $translate )
              if ($xml);
            while ( my $seq = $inseq->next_seq ) {
                $id++;
                if ( length( $seq->seq() ) > 20000 ) {
                    fasta_filter(
                        $seq, $ref,       undef,   undef,
                        $id,  $translate, $evalue, $xml,
                        $aug, $hmm,       $fs,     undef,
                        $verbose
                    );
                }
                else {
                    hmm_disc(
                        $seq,       $name,   $infile, $id,
                        $translate, $evalue, $xml,    $aug,
                        $hmm,       $fs,     undef,   $verbose
                    );
                }
            }
            Bio::Tools::ARFA->show('xml-end') if ($xml);
        }
        else {
            print "FATAL: Incorrect file format.\n";
        }
    }
    elsif ( defined $genbank ) {
        my ( $translate, $organism, $seq ) =
          Bio::Tools::ARFA->get_sequence($genbank);
        if ( defined $seq ) {
            $out--;
            Bio::Tools::ARFA->show( 'xml-begin', $evalue, $translate )
              if ($xml);
            if ( length( $seq->seq() ) > 20000 ) {
                fasta_filter(
                    $seq, $ref,       undef,   $organism,
                    1,    $translate, $evalue, $xml,
                    $aug, $hmm,       $fs,     undef,
                    $verbose
                );
            }
            else {
                hmm_disc(
                    $seq,       $name,   $infile, 1,
                    $translate, $evalue, $xml,    $aug,
                    $hmm,       $fs,     undef,   $verbose
                );
            }
            Bio::Tools::ARFA->show('xml-end') if ($xml);
        }
        else {
            $out += 2;
            print "  accession number not found!\n";
        }
    }
    else {
        $out++;
    }
}
else {
    print "\n..:: Automated Release Factor Annotation (ARFA) ::..\n> Standalone program version $version <\n\nFATAL: Incorrect arguments.\nUsage: rf_bot.pl [-options] -in=<sequence file>\n  Available options are:\n  -v      Print verbose output.\n  -x      XML output.\n  -table  Translation table number. (NCBI)\n  -name   Sequence name.\n  -e      Set the E-value threshold. It is a positive real number.\n          The default is 1e-40. Hit with E-value better than (less\n          than) this threshold will be shown.\n\n  Advanced options are:\n  --genbank=<ref>\n           Enable a search in GenBank.\n  -f       Force AUG only as start codon.\n\n";
    $out++;
}

exit($out);
